@@ -48,7 +48,6 @@ extern fontinfo *fonts;
 extern short fontcount;
 extern short attachto;
 extern char  _STR[150], _STR2[250];
-extern stringpart *labelbuf;
 extern int number_colors;
 
 extern double atan2();
@@ -62,6 +61,272 @@ char extchar[20];
 double saveratio;
 unsigned char texttype;
 
+/*--------------------------------------*/
+/* Element constructor functions	*/
+/*--------------------------------------*/
+
+/*--------------------------------------------------------------*/
+/* Label constructor:  Create a new label element in the object	*/
+/* whose instance is "destinst" and return a pointer to it.	*/
+/*								*/
+/*	"destinst" is the destination instance.  If NULL, the	*/
+/*	   top-level instance (areastruct.topinstance) is used.	*/
+/*	"strptr" is a pointer to a stringpart string, and may	*/
+/*	   be NULL.  If non-NULL, should NOT be free'd by the	*/
+/*	   calling routine.					*/
+/*	"pintype" is NORMAL, LOCAL, GLOBAL, or INFO		*/
+/*	"x" and "y" are the label coordinates.			*/
+/*								*/
+/* Other properties must be set individually by the calling	*/
+/* 	routine.						*/
+/*								*/
+/* Return value is a pointer to the newly created label.	*/
+/*--------------------------------------------------------------*/
+
+labelptr new_label(objinstptr destinst, stringpart *strptr, int pintype,
+	int x, int y)
+{
+   labelptr *newlab;
+   objectptr destobject;
+   objinstptr locdestinst;
+
+   locdestinst = (destinst == NULL) ? areastruct.topinstance : destinst;
+   destobject = locdestinst->thisobject;
+
+   NEW_LABEL(newlab, destobject);
+   destobject->parts++;
+   labeldefaults(*newlab, pintype, x, y);
+
+   if (strptr->type == FONT_NAME) {
+      free ((*newlab)->string);
+      (*newlab)->string = strptr;
+   }
+   else
+      (*newlab)->string->nextpart = strptr;
+
+   calcbboxvalues(locdestinst, (genericptr *)newlab);
+   updatepagebounds(destobject);
+   incr_changes(destobject);
+   return *newlab;
+}
+
+/*--------------------------------------------------------------*/
+/* Variant of the above; creates a label from a (char *) string	*/
+/* instead of a stringpart pointer.  Like the stringpart 	*/
+/* pointer above, "cstr" should NOT be free'd by the calling	*/
+/* routine.							*/
+/*--------------------------------------------------------------*/
+
+labelptr new_simple_label(objinstptr destinst, char *cstr,
+	int pintype, int x, int y)
+{
+   stringpart *strptr;
+
+   strptr = (stringpart *)malloc(sizeof(stringpart));
+   strptr->type = TEXT_STRING;
+   strptr->nextpart = NULL;
+   strptr->data.string = cstr;
+
+   return new_label(destinst, strptr, pintype, x, y);
+}
+
+/*--------------------------------------------------------------*/
+/* Another variant of the above; creates a "temporary" label	*/
+/* from a (char *) string.  As above, "cstr" should NOT be	*/
+/* free'd by the calling routine.  The "temporary" label has no	*/
+/* font information, and cannot be displayed nor saved/loaded	*/
+/* from the PostScript output file.  Used to name networks or	*/
+/* to mark port positions.  Pin type is always LOCAL.  Does not	*/
+/* require updating bounding box info since it cannot be	*/
+/* displayed.  Consequently, only requires passing the object	*/
+/* to get the new label, not its instance.			*/
+/*--------------------------------------------------------------*/
+
+labelptr new_temporary_label(objectptr destobject, char *cstr,
+	int x, int y)
+{
+   labelptr *newlab;
+
+   NEW_LABEL(newlab, destobject);
+   destobject->parts++;
+   labeldefaults(*newlab, LOCAL, x, y);
+
+   (*newlab)->string->type = TEXT_STRING; /* overwrites FONT record */
+   (*newlab)->string->data.string = cstr;
+   return *newlab;
+}
+
+/*--------------------------------------------------------------*/
+/* Polygon constructor:  Create a new polygon element in the	*/
+/* object whose instance is "destinst" and return a pointer to  */
+/* it.								*/
+/*								*/
+/*	"destinst" is the destination instance.  If NULL, the	*/
+/*	   top-level instance (areastruct.topinstance) is used.	*/
+/*	"points" is a list of XPoint pointers, should not be	*/
+/*	   NULL.  It is transferred to the polygon verbatim,	*/
+/*	   and should NOT be free'd by the calling routine.	*/
+/*	"number" is the number of points in the list, or zero	*/
+/*	   if "points" is NULL.					*/
+/*								*/
+/* Other properties must be set individually by the calling	*/
+/* 	routine.						*/
+/*								*/
+/* Return value is a pointer to the newly created polygon.	*/
+/*--------------------------------------------------------------*/
+
+polyptr new_polygon(objinstptr destinst, pointlist *points, int number)
+{
+   polyptr *newpoly;
+   objectptr destobject;
+   objinstptr locdestinst;
+
+   locdestinst = (destinst == NULL) ? areastruct.topinstance : destinst;
+   destobject = locdestinst->thisobject;
+
+   NEW_POLY(newpoly, destobject);
+   destobject->parts++;
+   polydefaults(*newpoly, 0, 0, 0);
+   (*newpoly)->number = number;
+   (*newpoly)->points = *points;
+
+   calcbboxvalues(locdestinst, (genericptr *)newpoly);
+   updatepagebounds(destobject);
+   incr_changes(destobject);
+   return *newpoly;
+}
+
+/*--------------------------------------------------------------*/
+/* Spline constructor:  Create a new spline element in the	*/
+/* object whose instance is "destinst" and return a pointer to  */
+/* it.								*/
+/*								*/
+/*	"destinst" is the destination instance.  If NULL, the	*/
+/*	   top-level instance (areastruct.topinstance) is used.	*/
+/*	"points" is a array of 4 XPoints; should not be	NULL.	*/
+/*								*/
+/* Other properties must be set individually by the calling	*/
+/* 	routine.						*/
+/*								*/
+/* Return value is a pointer to the newly created spline.	*/
+/*--------------------------------------------------------------*/
+
+splineptr new_spline(objinstptr destinst, pointlist points)
+{
+   splineptr *newspline;
+   objectptr destobject;
+   objinstptr locdestinst;
+   int i;
+
+   locdestinst = (destinst == NULL) ? areastruct.topinstance : destinst;
+   destobject = locdestinst->thisobject;
+
+   NEW_SPLINE(newspline, destobject);
+   destobject->parts++;
+   splinedefaults(*newspline, 0, 0);
+
+   for (i = 0; i < 4; i++)
+      (*newspline)->ctrl[i] = points[i];
+
+   calcspline(*newspline);
+   calcbboxvalues(locdestinst, (genericptr *)newspline);
+   updatepagebounds(destobject);
+   incr_changes(destobject);
+   return *newspline;
+}
+
+/*--------------------------------------------------------------*/
+/* Arc constructor:  Create a new arc element in the object	*/
+/* whose instance is "destinst" and return a pointer to it.	*/
+/*								*/
+/*	"destinst" is the destination instance.  If NULL, the	*/
+/*	   top-level instance (areastruct.topinstance) is used.	*/
+/*	"radius" is the radius of the (circular) arc.		*/
+/*	"x" and "y" represents the arc center position.		*/
+/*								*/
+/* Other properties must be set individually by the calling	*/
+/* 	routine.						*/
+/*								*/
+/* Return value is a pointer to the newly created arc.		*/
+/*--------------------------------------------------------------*/
+
+arcptr new_arc(objinstptr destinst, int radius, int x, int y)
+{
+   arcptr *newarc;
+   objectptr destobject;
+   objinstptr locdestinst;
+
+   locdestinst = (destinst == NULL) ? areastruct.topinstance : destinst;
+   destobject = locdestinst->thisobject;
+
+   NEW_ARC(newarc, destobject);
+   destobject->parts++;
+   arcdefaults(*newarc, x, y);
+   (*newarc)->radius = (*newarc)->yaxis = radius;
+
+   calcarc(*newarc);
+   calcbboxvalues(locdestinst, (genericptr *)newarc);
+   updatepagebounds(destobject);
+   incr_changes(destobject);
+   return *newarc;
+}
+
+/*--------------------------------------------------------------*/
+/* Instance constructor:  Create a new object instance element	*/
+/* in the object whose instance is "destinst" and return a	*/
+/* pointer to it.						*/
+/*								*/
+/*	"destinst" is the destination instance.  If NULL, the	*/
+/*	   top-level instance (areastruct.topinstance) is used.	*/
+/*	"srcinst" is the source instance of which this is a	*/
+/*	   copy.						*/
+/*	"x" and "y" represents the instance position.		*/
+/*								*/
+/* Other properties must be set individually by the calling	*/
+/* 	routine.						*/
+/*								*/
+/* Return value is a pointer to the newly created arc.		*/
+/*--------------------------------------------------------------*/
+
+objinstptr new_objinst(objinstptr destinst, objinstptr srcinst, int x, int y)
+{
+   objinstptr *newobjinst;
+   objectptr destobject;
+   objinstptr locdestinst;
+
+   locdestinst = (destinst == NULL) ? areastruct.topinstance : destinst;
+   destobject = locdestinst->thisobject;
+
+   NEW_OBJINST(newobjinst, destobject);
+   destobject->parts++;
+   instcopy(*newobjinst, srcinst);
+   (*newobjinst)->position.x = x;
+   (*newobjinst)->position.y = y;
+
+   calcbboxvalues(locdestinst, (genericptr *)newobjinst);
+   updatepagebounds(destobject);
+   incr_changes(destobject);
+   return *newobjinst;
+}
+
+/*--------------------------------------------------------------*/
+/* Generic element destructor function				*/
+/*--------------------------------------------------------------*/
+
+void remove_element(objinstptr destinst, genericptr genelem)
+{
+   objectptr destobject;
+   objinstptr locdestinst;
+
+   locdestinst = (destinst == NULL) ? areastruct.topinstance : destinst;
+   destobject = locdestinst->thisobject;
+
+   genelem->type &= REMOVE_TAG;
+   delete_tagged(destobject);
+   calcbboxvalues(locdestinst, (genericptr *)NULL);
+   updatepagebounds(destobject);
+}
+
 /*-------------------------------------*/
 /* Sane values for a new path instance */
 /*-------------------------------------*/
@@ -81,7 +346,7 @@ void pathdefaults(pathptr newpath, int x, int y)
 /* Sane values for a new object instance */
 /*---------------------------------------*/
 
-void objectdefaults(objinstptr newinst, objectptr thisobj, int x, int y)
+void instancedefaults(objinstptr newinst, objectptr thisobj, int x, int y)
 {
    newinst->position.x = x;
    newinst->position.y = y;
@@ -97,9 +362,7 @@ void objectdefaults(objinstptr newinst, objectptr thisobj, int x, int y)
    newinst->bbox.width = thisobj->bbox.width;
    newinst->bbox.height = thisobj->bbox.height;
 
-#ifdef SCHEMA
    newinst->schembbox = NULL;
-#endif
 }
 
 /*--------------------------------------*/
@@ -117,7 +380,8 @@ void drawdot(int xpos, int ypos)
    if ((dotobj = finddot()) != (objectptr)NULL) {
       NEW_OBJINST(newdot, topobject);
       topobject->parts++;
-      objectdefaults(*newdot, dotobj, xpos, ypos);
+      instancedefaults(*newdot, dotobj, xpos, ypos);
+      register_for_undo(XCF_Dot, UNDO_DONE, areastruct.topinstance, *newdot);
    }
    else {
       NEW_ARC(newarc, topobject);
@@ -129,6 +393,7 @@ void drawdot(int xpos, int ypos)
       (*newarc)->style = FILLED | FILLSOLID | NOBORDER;
       (*newarc)->passed = NULL;
       calcarc(*newarc);
+      register_for_undo(XCF_Arc, UNDO_DONE, areastruct.topinstance, *newarc);
    }
    incr_changes(topobject);
 }
@@ -175,14 +440,13 @@ void labeldefaults(labelptr newlabel, u_char dopin, int x, int y)
    newlabel->string->data.font = areastruct.psfont;
    newlabel->string->nextpart = NULL;
 
-#ifdef SCHEMA
    newlabel->pin = False;
    if (areastruct.schemon) {
       newlabel->pin = dopin;
       if (dopin == LOCAL) newlabel->color = LOCALPINCOLOR;
+      else if (dopin == GLOBAL) newlabel->color = GLOBALPINCOLOR;
       else if (dopin == INFO) newlabel->color = INFOLABELCOLOR;
    }
-#endif
 
    newlabel->justify = areastruct.justify;
    newlabel->position.x = x;
@@ -258,7 +522,7 @@ labelptr findlabelcopy(labelptr curlabel, stringpart *curstring)
    labelptr tlab;
 
    for (tgen = topobject->plist; tgen < topobject->plist + topobject->parts; tgen++) {
-      if ((*tgen)->type == LABEL) {
+      if (IS_LABEL(*tgen)) {
          tlab = TOLABEL(tgen);
 	 if (tlab->pin == False) continue;
 	 else if (tlab == curlabel) continue;  /* Don't count self! */
@@ -277,7 +541,7 @@ labelptr findlabelcopy(labelptr curlabel, stringpart *curstring)
 void labeltext(int keypressed, char *clientdata)
 {
    labelptr curlabel;
-   stringpart *curpos;
+   stringpart *curpos, *labelbuf;
    int locpos;
    Boolean do_redraw = False;
    short tmplength, tmpheight, cfont;
@@ -343,20 +607,17 @@ void labeltext(int keypressed, char *clientdata)
 
       if (hasstuff && (eventmode != TEXT3_MODE && eventmode != CATTEXT_MODE)) {
 	 topobject->parts++;
+         register_for_undo(XCF_Text, UNDO_DONE, areastruct.topinstance,
+                    curlabel);
+
 	 incr_changes(topobject);
-#ifdef SCHEMA
 	 invalidate_netlist(topobject);
-#endif
 
       }
       else if (!hasstuff && (eventmode == TEXT3_MODE)) topobject->parts--;
 
       if ((!hasstuff) && (eventmode == CATTEXT_MODE)) {  /* can't have null labels! */ 
-	  freelabel(curlabel->string);
-	  curlabel->string = stringcopyback(labelbuf, areastruct.topinstance);
-	  freelabel(labelbuf);
-	  labelbuf = NULL;
-	  resolveparams(areastruct.topinstance);
+	  undo_action();
           XcSetFunction(GXcopy);
 	  redrawtext(curlabel);
 	  Wprintf("Object must have a name!");
@@ -373,7 +634,7 @@ void labeltext(int keypressed, char *clientdata)
 
 	 areastruct.save.y = curlabel->position.y + 100;
 	 areastruct.save.x = curlabel->position.x;
-	 objectselect(OBJECT);
+	 select_element(OBJINST);
 	 if (areastruct.selects == 1) {  /* this should be true! */
 	    short *selptr = areastruct.selectlist + areastruct.selects - 1;
 	    objinstptr selobj = SELTOOBJINST(selptr);
@@ -393,21 +654,18 @@ void labeltext(int keypressed, char *clientdata)
 	    }
 	 }
          eventmode = CATALOG_MODE;
-	 objectdeselect();
+	 unselect_all();
       }
       else {	/* (hasstuff && eventmode != CATTEXT_MODE) */
 	 eventmode = NORMAL_MODE;
 	 incr_changes(topobject);
-#ifdef SCHEMA
 	 if (curlabel->pin != False) invalidate_netlist(topobject);
-#endif
       }
 
       setdefaultfontmarks();
       setcolormark(areastruct.color);
-      if (labelbuf != NULL) {
+      if ((labelbuf = get_original_string(curlabel)) != NULL) {
 
-#ifdef SCHEMA
 	 /* If the original label (before modification) is a pin in a	*/
 	 /* schematic/symbol with a matching symbol/schematic, and the	*/
 	 /* name is unique, change every pin in the matching symbol/	*/
@@ -426,15 +684,10 @@ void labeltext(int keypressed, char *clientdata)
 	       invalidate_netlist(topobject->symschem);
 	    }
 	 }
-#endif
 	
-	 freelabel(labelbuf);
-	 labelbuf = NULL;
 	 resolveparams(areastruct.topinstance);
-         updateinstparam(areastruct.topinstance);
-#ifdef SCHEMA
+         updateinstparam(topobject);
 	 setobjecttype(topobject);
-#endif
       }
       else
          calcbbox(areastruct.topinstance);
@@ -619,11 +872,9 @@ void labeltext(int keypressed, char *clientdata)
       areastruct.origin.y = curlabel->position.y + (curlabel->justify &
 	  NOTBOTTOM ? (curlabel->justify & TOP ? -tmpheight : -tmpheight / 2)
 	  : 0);
-#ifdef SCHEMA
       if (curlabel->pin)
          pinadjust(curlabel->justify, &(areastruct.origin.x),
 		&(areastruct.origin.y), 1);
-#endif
       XcSetFunction(GXcopy);
       redrawtext(curlabel);
    }
@@ -696,7 +947,7 @@ void rejustify(short mode)
 	    redrawtext(curlabel);
 	 }
       }
-      if (eventmode != PRESS_MODE && eventmode != COPY2_MODE) objectdeselect();
+      if (eventmode != PRESS_MODE && eventmode != COPY2_MODE) unselect_all();
    }
    if (curlabel == NULL)
       Wprintf("No labels chosen to rejustify");
@@ -948,12 +1199,17 @@ void polydefaults(polyptr newpoly, int number, int x, int y)
    newpoly->color = areastruct.color;
    newpoly->width = areastruct.linewidth;
    newpoly->number = number;
-   newpoly->points = (pointlist) malloc(number * sizeof(XPoint));
    newpoly->passed = NULL;
-   for (pointptr = newpoly->points; pointptr < newpoly->points + number;
-	pointptr++) {
-      pointptr->x = x;
-      pointptr->y = y;
+   if (number == 0)
+      newpoly->points = NULL;
+   else {
+      newpoly->points = (pointlist) malloc(number * sizeof(XPoint));
+
+      for (pointptr = newpoly->points; pointptr < newpoly->points + number;
+		pointptr++) {
+         pointptr->x = x;
+         pointptr->y = y;
+      }
    }
 }
 
@@ -1218,7 +1474,7 @@ void setendpoint(short *scnt, short direc, XPoint **endpoint, XPoint *arcpoint)
 {
    genericptr *sptr = topobject->plist + (*scnt);
 
-   switch((*sptr)->type) {
+   switch(ELEMENTTYPE(*sptr)) {
       case POLYGON:
 	 if (direc)
 	    *endpoint = TOPOLY(sptr)->points + TOPOLY(sptr)->number - 1;
@@ -1292,19 +1548,13 @@ void reversefpoints(XfPoint *plist, short number)
 /*	add = 1 if plist has (parts + 1) elements		*/
 /*--------------------------------------------------------------*/
 
-void freeparts(short *selectobj, short add)
+void freepathparts(short *selectobj, short add)
 {
    genericptr *oldelem = topobject->plist + (*selectobj);
-   switch((*oldelem)->type) {
+   switch(ELEMENTTYPE(*oldelem)) {
       case POLYGON:
 	 free((TOPOLY(oldelem))->points);
 	 break;
-      case PATH:
-	 free((TOPATH(oldelem))->plist);
-	 break;
-      case LABEL:
-	 free((TOLABEL(oldelem))->string);
-	 break;
    }
    free(*oldelem);
    removep(selectobj, add);
@@ -1335,7 +1585,7 @@ void unjoin()
    genericptr *genp, *newg;
    pathptr oldpath;
 
-   if (areastruct.selects == 0) objectselect(PATH);
+   if (areastruct.selects == 0) select_element(PATH);
    if (areastruct.selects == 0) {
       Wprintf("No objects selected.");
       return;
@@ -1368,8 +1618,8 @@ void unjoin()
 
          /* remove the path object and revise the selectlist */
 
-         freeparts(selectobj, 0);
-         reviseselect(selectobj);
+         freepathparts(selectobj, 0);
+         reviseselect(areastruct.selectlist, areastruct.selects, selectobj);
       }
    }
    clearselects();
@@ -1595,7 +1845,7 @@ void join()
 
       for (sptr = scount; sptr < scount + numpolys; sptr++) {
 	 easydraw(*sptr, DOFORALL);
-	 freeparts(sptr, 1);
+	 freepathparts(sptr, 1);
 	 /* revise the list of path elements */
 	 for (sptr2 = sptr + 1; sptr2 < scount + numpolys; sptr2++)
 	    if (*sptr2 > *sptr) (*sptr2)--;
@@ -1624,7 +1874,7 @@ void join()
 	 /* reverse point order if necessary */
 
          if (*(direc + (short)(sptr - order)) == 1) {
-	    switch ((*newelem)->type) {
+	    switch (ELEMENTTYPE(*newelem)) {
 	       case POLYGON:
 		  reversepoints(TOPOLY(newelem)->points, TOPOLY(newelem)->number);
 	          break;
@@ -1728,9 +1978,8 @@ void wirebutton(XButtonEvent *event)
          XcSetForeground(newwire->color);
 	 topobject->parts++;
 	 incr_changes(topobject);
-#ifdef SCHEMA
 	 if (!nonnetwork(newwire)) invalidate_netlist(topobject);
-#endif
+	 register_for_undo(XCF_Wire, UNDO_DONE, areastruct.topinstance, newwire);
       }
       UDrawPolygon(newwire);
       if (event->button == Button3)
