@@ -291,23 +291,44 @@ float findsplinemin(splineptr thespline, XPoint *upoint)
 /* Find closest point of a polygon to the cursor			      */
 /*----------------------------------------------------------------------------*/
 
-short closepoint(polyptr curpoly, XPoint *cursloc)
+short closepointdistance(polyptr curpoly, XPoint *cursloc, short *mindist)
 {
-   short curdist, mindist;
+   short curdist;
    XPoint *curpt, *savept; 
 
    curpt = savept = curpoly->points;
-   mindist = wirelength(curpt, cursloc);
+   *mindist = wirelength(curpt, cursloc);
    while (++curpt < curpoly->points + curpoly->number) {
       curdist = wirelength(curpt, cursloc);
-      if (curdist < mindist) {
-         mindist = curdist;
+      if (curdist < *mindist) {
+         *mindist = curdist;
          savept = curpt;
       }
    }
    return (short)(savept - curpoly->points);
 }
 
+/*----------------------------------------------------------------------------*/
+/* Find closest point of a polygon to the cursor			      */
+/*----------------------------------------------------------------------------*/
+
+short closepoint(polyptr curpoly, XPoint *cursloc)
+{
+   short mindist;
+   return closepointdistance(curpoly, cursloc, &mindist);
+}
+
+/*----------------------------------------------------------------------------*/
+/* Find the distance to the closest point of a polygon to the cursor	      */
+/*----------------------------------------------------------------------------*/
+
+short closedistance(polyptr curpoly, XPoint *cursloc)
+{
+   short mindist;
+   closepointdistance(curpoly, cursloc, &mindist);
+   return mindist;
+}
+
 /*----------------------------------------------------------------------------*/
 /* Coordinate system transformations 					      */
 /*----------------------------------------------------------------------------*/
@@ -857,14 +878,17 @@ void calcextents(genericptr *bboxgen, short *llx, short *lly,
 /* Calculate the bounding box of an object instance			*/
 /*----------------------------------------------------------------------*/
 
-void objinstbbox(objinstptr obbox, XPoint *npoints)
+void objinstbbox(objinstptr obbox, XPoint *npoints, Boolean expand)
 {
    XPoint points[4];
+   int extend = (expand) ? 4 : 0;
 
-   points[0].x = points[1].x = obbox->bbox.lowerleft.x;
-   points[1].y = points[2].y = obbox->bbox.lowerleft.y + obbox->bbox.height;
-   points[2].x = points[3].x = obbox->bbox.lowerleft.x + obbox->bbox.width;
-   points[0].y = points[3].y = obbox->bbox.lowerleft.y;
+   points[0].x = points[1].x = obbox->bbox.lowerleft.x - extend;
+   points[1].y = points[2].y = obbox->bbox.lowerleft.y + obbox->bbox.height
+		+ extend;
+   points[2].x = points[3].x = obbox->bbox.lowerleft.x + obbox->bbox.width
+		+ extend;
+   points[0].y = points[3].y = obbox->bbox.lowerleft.y - extend;
 
    UTransformPoints(points, npoints, 4, obbox->position,
 		 obbox->scale, obbox->rotation); 
@@ -930,7 +954,7 @@ void calcbboxsingle(genericptr *bboxgen, objinstptr thisinst,
    switch(ELEMENTTYPE(*bboxgen)) {
 
       case(OBJINST):
-	 objinstbbox(TOOBJINST(bboxgen), npoints);
+	 objinstbbox(TOOBJINST(bboxgen), npoints, False);
 
          for (j = 0; j < 4; j++) {
             bboxcalc(npoints[j].x, llx, urx);
