@@ -445,13 +445,14 @@ stringpart *PySetStringParts(PyObject *dval)
 	 }
 	 else if ((ditem = PyDict_GetItemString(litem, "Text")) != NULL) {
 	    newpart->type = TEXT_STRING;
-	    newpart->data.string = mystrdup(PyString_AsString(ditem));
+	    newpart->data.string = strdup(PyString_AsString(ditem));
 	 }
       }
       else if (PyString_Check(litem)) {
 	 string = PyString_AsString(litem);
 	 if (!strcmp(string, "End Parameter")) {
 	    newpart->type = PARAM_END;
+	    newpart->data.string = (u_char *)NULL;
 	 }
 	 else if (!strcmp(string, "Tab Stop")) {
 	    newpart->type = TABSTOP;
@@ -491,7 +492,7 @@ stringpart *PySetStringParts(PyObject *dval)
 	 }
 	 else {
 	    newpart->type = TEXT_STRING;
-	    newpart->data.string = mystrdup(string);
+	    newpart->data.string = strdup(string);
 	 }
       }
    }
@@ -885,7 +886,6 @@ static PyObject *xc_reset()
 /* Get a netlist						*/
 /*--------------------------------------------------------------*/
 
-extern u_int subindex;  /* count for labeling spice subcircuits */
 extern PyObject *pyglobals(objectptr);
 extern PyObject *pytoplevel(objectptr);
 
@@ -898,7 +898,6 @@ static PyObject *xc_netlist()
       return NULL;
    }
 
-   subindex = 1;
    rdict = PyDict_New();
    PyDict_SetItem(rdict, PyString_FromString("globals"), pyglobals(topobject));
    PyDict_SetItem(rdict, PyString_FromString("circuit"), pytoplevel(topobject));
